/*
 * swmod.c: SWMOD deamon
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
 
#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <libubox/uloop.h>

#include "swmod_uci.h"
#include "swmod_opkg.h"
#include "swmod_api.h"
#include "tools.h"
#include "swmod_host.h"
#include "swmod.h"
#ifdef SWMOD_LXC
#include "swmod_lxc.h"
#endif

struct ubus_context *ubus_ctx;
ExecEnv_t g_environments[MAX_ENV] = {0};

enum {
	DU_INSTALL_ENV_ID,
	DU_INSTALL_ENV,
	DU_INSTALL_UUID,
	DU_INSTALL_URL,
	DU_INSTALL_USERNAME,
	DU_INSTALL_PASSWORD,
	__DU_INSTALL_MAX
};

enum {
	DU_UPDATE_ENV_ID,
	DU_UPDATE_ENV,
	DU_UPDATE_UUID,
	DU_UPDATE_URL,
	DU_UPDATE_USERNAME,
	DU_UPDATE_PASSWORD,
	__DU_UPDATE_MAX
};

enum {
	DU_UNINSTALL_ENV_ID,
	DU_UNINSTALL_ENV,
	DU_UNINSTALL_NAME,
	__DU_UNINSTALL_MAX
};

enum {
	EU_DU_LIST_ENV_ID,
	EU_DU_LIST_ENV,
	__EU_DU_LIST_MAX
};

enum {
	EU_SET_STATE_SERVICE_ENV_ID,
	EU_SET_STATE_SERVICE_ENV,
	EU_SET_STATE_SERVICE_NAME,
	EU_SET_STATE_SERVICE_STATUS,
	__EU_SET_STATE_SERVICE_MAX
};

enum {
	EE_SET_STATE_ENV_ID,
	EE_SET_STATE_ENV,
	EE_SET_STATE_STATUS,
	__EE_SET_STATE_MAX
};

static const struct blobmsg_policy ee_set_state_policy[__EE_SET_STATE_MAX] = {
	[EE_SET_STATE_ENV_ID] = { .name = "eeid", .type = BLOBMSG_TYPE_INT32 },
	[EE_SET_STATE_ENV] = { .name = "ee_name", .type = BLOBMSG_TYPE_STRING },
	[EE_SET_STATE_STATUS] = { .name = "state", .type = BLOBMSG_TYPE_STRING },
};

static const struct blobmsg_policy eu_set_state_policy[__EU_SET_STATE_SERVICE_MAX] = {
	[EU_SET_STATE_SERVICE_ENV_ID] = { .name = "eeid", .type = BLOBMSG_TYPE_INT32 },
	[EU_SET_STATE_SERVICE_ENV] = { .name = "ee_name", .type = BLOBMSG_TYPE_STRING },
	[EU_SET_STATE_SERVICE_NAME] = { .name = "eu_name", .type = BLOBMSG_TYPE_STRING },
	[EU_SET_STATE_SERVICE_STATUS] = { .name = "state", .type = BLOBMSG_TYPE_INT8 },
};

static const struct blobmsg_policy du_install_policy[__DU_INSTALL_MAX] = {
	[DU_INSTALL_ENV_ID] = { .name = "eeid", .type = BLOBMSG_TYPE_INT32 },
	[DU_INSTALL_ENV] = { .name = "ee_name", .type = BLOBMSG_TYPE_STRING },
	[DU_INSTALL_UUID] = { .name = "uuid", .type = BLOBMSG_TYPE_STRING },
	[DU_INSTALL_URL] = { .name = "url", .type = BLOBMSG_TYPE_STRING },
	[DU_INSTALL_USERNAME] = { .name = "username", .type = BLOBMSG_TYPE_STRING },
	[DU_INSTALL_PASSWORD] = { .name = "password", .type = BLOBMSG_TYPE_STRING },
};

static const struct blobmsg_policy du_update_policy[__DU_UPDATE_MAX] = {
	[DU_UPDATE_ENV_ID] = { .name = "eeid", .type = BLOBMSG_TYPE_INT32 },
	[DU_UPDATE_ENV] = { .name = "ee_name", .type = BLOBMSG_TYPE_STRING },
	[DU_UPDATE_UUID] = { .name = "uuid", .type = BLOBMSG_TYPE_STRING },
	[DU_UPDATE_URL] = { .name = "url", .type = BLOBMSG_TYPE_STRING },
	[DU_UPDATE_USERNAME] = { .name = "username", .type = BLOBMSG_TYPE_STRING },
	[DU_UPDATE_PASSWORD] = { .name = "password", .type = BLOBMSG_TYPE_STRING },
};

static const struct blobmsg_policy du_uninstall_policy[__DU_UNINSTALL_MAX] = {
	[DU_UNINSTALL_ENV_ID] = { .name = "eeid", .type = BLOBMSG_TYPE_INT32 },
	[DU_UNINSTALL_ENV] = { .name = "ee_name", .type = BLOBMSG_TYPE_STRING },
	[DU_UNINSTALL_NAME] = { .name = "du_name", .type = BLOBMSG_TYPE_STRING },
};

static const struct blobmsg_policy eu_du_list_policy[__EU_DU_LIST_MAX] = {
	[EU_DU_LIST_ENV_ID] = { .name = "eeid", .type = BLOBMSG_TYPE_INT32 },
	[EU_DU_LIST_ENV] = { .name = "ee_name", .type = BLOBMSG_TYPE_STRING },
};

static void
populate_environments(void)
{
	memset(g_environments, '\0', sizeof(g_environments));

	/* Host system */
#ifdef SWMOD_HOST_IS_EXECENV
	populate_host_system_environment();
#endif

	/* Linux containers */
#ifdef SWMOD_LXC
	populate_lxc_environment();
#endif
}

static ExecEnv_t* get_ee_from_eeid(int eid)
{
	int i;
	for (i = 0; i < MAX_ENV; i++) {
		if (!g_environments[i].exists)
			continue;

		if (eid == g_environments[i].eeid) {
			return &g_environments[i];
		}
	}

	return NULL;
}

static ExecEnv_t* get_ee_from_name(char *ename)
{
	if (ename == NULL)
		return NULL;

	int i;
	for (i = 0; i < MAX_ENV; i++) {
		if (!g_environments[i].exists)
			continue;

		if (strcmp(g_environments[i].name, ename) == 0) {
			return &g_environments[i];
		}
	}

	return NULL;
}

static int
swmod_environment(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_buf bb;
	void *a, *t;
	int i;

	populate_environments();

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	a = blobmsg_open_array(&bb, "environment");

	for (i = 0; i < MAX_ENV; i++) {
		if (!g_environments[i].exists)
			continue;

		t = blobmsg_open_table(&bb, "");

		blobmsg_add_string(&bb, "ee_name", g_environments[i].name);
		blobmsg_add_u32(&bb, "eeid", g_environments[i].eeid);
		blobmsg_add_string(&bb, "status", g_environments[i].status);
		blobmsg_add_u32(&bb, "pause", g_environments[i].pause);
		blobmsg_add_u32(&bb, "autoboot", g_environments[i].autoboot);
		blobmsg_add_string(&bb, "type", g_environments[i].type);
		blobmsg_add_string(&bb, "vendor", g_environments[i].vendor);
		blobmsg_add_string(&bb, "version", g_environments[i].version);
		blobmsg_add_u64(&bb, "allocated_disk_space", g_environments[i].allocated_disk_space);
		blobmsg_add_u64(&bb, "available_disk_space", g_environments[i].available_disk_space);
		blobmsg_add_u64(&bb, "allocated_memory", g_environments[i].allocated_memory);
		blobmsg_add_u64(&bb, "available_memory", g_environments[i].available_memory);
		blobmsg_add_u32(&bb, "parent_ee_ref", g_environments[i].parent_eeid);

		blobmsg_close_table(&bb, t);
	}

	blobmsg_close_array(&bb, a);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);
	return 0;
}

static int
swmod_du_install(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_attr *tb[__DU_INSTALL_MAX] = {NULL};
	struct blob_buf bb;
	int err = -1;
	ExecEnv_t *env = NULL;
	PkgInfo pkg;

	memset(&pkg, 0, sizeof(pkg));

	if (blobmsg_parse(du_install_policy, __DU_INSTALL_MAX, tb, blob_data(msg), blob_len(msg)))
		return UBUS_STATUS_UNKNOWN_ERROR;

	if (tb[DU_INSTALL_ENV])
		env = get_ee_from_name(blobmsg_get_string(tb[DU_INSTALL_ENV]));

	/* Priority should be given on EID if both EID and ENV name are provided */
	if (tb[DU_INSTALL_ENV_ID])
		env = get_ee_from_eeid(blobmsg_get_u32(tb[DU_INSTALL_ENV_ID]));

	if (env == NULL)
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (!tb[DU_INSTALL_URL])
		return UBUS_STATUS_INVALID_ARGUMENT;

	snprintf(pkg.url, sizeof(pkg.url), "%s", blobmsg_get_string(tb[DU_INSTALL_URL]));

	if (tb[DU_INSTALL_USERNAME])
		snprintf(pkg.uname, sizeof(pkg.uname), "%s", blobmsg_get_string(tb[DU_INSTALL_USERNAME]));

	if (tb[DU_INSTALL_PASSWORD])
		snprintf(pkg.psw, sizeof(pkg.psw), "%s", blobmsg_get_string(tb[DU_INSTALL_PASSWORD]));

	char output[MAX_LEN_1024] = {0};
	time_t start = time(NULL);

	err = swmod_install_package(env, &pkg, output, sizeof(output));

	char package_name[MAX_LEN_64], package_version[MAX_LEN_64];
	memset(package_name, 0, sizeof(package_name));
	memset(package_version, 0, sizeof(package_version));
	if (err == 0) {
		if (strstr(output, "Installing") == NULL) {
			err = -1;
		} else {
			sscanf(output, "Installing %63s (%63[^)] to root...", package_name, package_version);
		}
	}

	bool failed = (err) ? true : false;
	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);
	blobmsg_add_u8(&bb, "status", !failed);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	time_t finish = time(NULL);
	swmod_send_event_du_state_change(failed, start, finish, package_version, SWMOD_INSTALL, output);
	if (failed == false) {
		synchronize_du_list(env);
	}

	return 0;
}

static char *get_performed_operation(int action)
{
	switch(action) {
	case SWMOD_INSTALL:
		return "Install";
	case SWMOD_UPDATE:
		return "Update";
	case SWMOD_REMOVE:
		return "Uninstall";
	}

	return "Invalid";
}

static char *get_current_du_state(int action, bool failed)
{
	switch(action) {
	case SWMOD_INSTALL:
		if (failed)
			return "Failed";

		return "Installed";
	case SWMOD_UPDATE:
		return "Installed";
	case SWMOD_REMOVE:
		if (failed)
			return "Installed";
		return "UnInstalled";
	}

	return "Invalid";
}

void swmod_send_event_du_state_change(bool failed, time_t start, time_t end, char *ver, int action, char *err)
{
	uint32_t id;
	char start_time[MAX_LEN_32] = {0};
	char end_time[MAX_LEN_32] = {0};

	if (ubus_ctx == NULL) {
		return;
	}

	strftime(start_time, sizeof(start_time), "%Y-%m-%dT%H:%M:%SZ", gmtime(&start));
	strftime(end_time, sizeof(end_time), "%Y-%m-%dT%H:%M:%SZ", gmtime(&end));
	char *operation = get_performed_operation(action);
	char *state = get_current_du_state(action, failed);

	struct blob_buf b;
	memset(&b, 0, sizeof(struct blob_buf));

	blob_buf_init(&b, 0);
	blobmsg_add_string(&b, "name", "Device.SoftwareModules.DUStateChange!");

	void *tbl = blobmsg_open_table(&b, "input");
	blobmsg_add_string(&b, "Version", ver);
	blobmsg_add_string(&b, "CurrentState", state);
	blobmsg_add_string(&b, "Resolved", failed ? "0" : "1");
	blobmsg_add_string(&b, "StartTime", start_time);
	blobmsg_add_string(&b, "CompleteTime", end_time);
	blobmsg_add_string(&b, "OperationPerformed", operation);

	if (failed) {
		blobmsg_add_string(&b, "Fault.FaultCode", "9001");
		if (err && err[0] != '\0')
			blobmsg_add_string(&b, "Fault.FaultString", err);
		else
			blobmsg_add_string(&b, "Fault.FaultString", "Internal Error");
	} else {
		blobmsg_add_string(&b, "Fault.FaultCode", "0");
		blobmsg_add_string(&b, "Fault.FaultString", "");
	}

	blobmsg_close_table(&b, tbl);

	if (!ubus_lookup_id(ubus_ctx, "usp.raw", &id))
		ubus_invoke(ubus_ctx, id, "notify_event", b.head, NULL, NULL, 5);

	blob_buf_free(&b);

	return;
}

static int
swmod_du_update(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_attr *tb[__DU_UPDATE_MAX] = {NULL};
	struct blob_buf bb;
	int err = -1;
	ExecEnv_t *env = NULL;
	PkgInfo pkg;

	memset(&pkg, 0, sizeof(pkg));

	if (blobmsg_parse(du_update_policy, __DU_UPDATE_MAX, tb, blob_data(msg), blob_len(msg)))
		return UBUS_STATUS_UNKNOWN_ERROR;

	if (!tb[DU_UPDATE_URL] || !tb[DU_UPDATE_UUID])
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (tb[DU_UPDATE_ENV])
		env = get_ee_from_name(blobmsg_get_string(tb[DU_UPDATE_ENV]));

	/* Priority should be given on EID if both EID and ENV name are provided */
	if (tb[DU_UPDATE_ENV_ID])
		env = get_ee_from_eeid(blobmsg_get_u32(tb[DU_UPDATE_ENV_ID]));

	if (env == NULL)
		return UBUS_STATUS_INVALID_ARGUMENT;

	snprintf(pkg.url, sizeof(pkg.url), "%s", blobmsg_get_string(tb[DU_UPDATE_URL]));

	if (tb[DU_UPDATE_USERNAME])
		snprintf(pkg.uname, sizeof(pkg.uname), "%s", blobmsg_get_string(tb[DU_UPDATE_USERNAME]));

	if (tb[DU_UPDATE_PASSWORD])
		snprintf(pkg.psw, sizeof(pkg.psw), "%s", blobmsg_get_string(tb[DU_UPDATE_PASSWORD]));

	char output[MAX_LEN_1024] = {0};
	time_t start = time(NULL);

	err = swmod_update_package(env, &pkg, output, sizeof(output));

	char package_name[MAX_LEN_64], package_version[MAX_LEN_64];
	memset(package_name, 0, sizeof(package_name));
	memset(package_version, 0, sizeof(package_version));
	if (err == 0) {
		if (strstr(output, "Upgrading") == NULL) {
			err = -1;
		} else {
			char pkg_old_version[MAX_LEN_64] = {0};
			sscanf(output, "Upgrading %63s on root from %63s to %63s", package_name, pkg_old_version, package_version);
			int len = strlen(package_version);
			if (len > 3 && package_version[len - 3] == '.')
				package_version[len - 3] = '\0';
		}
	}

	bool failed = (err) ? true : false;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);
	blobmsg_add_u8(&bb, "status", !failed);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	time_t finish = time(NULL);
	swmod_send_event_du_state_change(failed, start, finish, package_version, SWMOD_UPDATE, output);
	if (failed == false) {
		synchronize_du_list(env);
	}

	return 0;
}

static int
swmod_du_uninstall(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_attr *tb[__DU_UNINSTALL_MAX] = {NULL};
	struct blob_buf bb;
	ExecEnv_t *env = NULL;
	PkgInfo pkg;

	memset(&pkg, 0, sizeof(pkg));

	if (blobmsg_parse(du_uninstall_policy, __DU_UNINSTALL_MAX, tb, blob_data(msg), blob_len(msg)))
		return UBUS_STATUS_UNKNOWN_ERROR;

	if (!tb[DU_UNINSTALL_NAME])
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (tb[DU_UNINSTALL_ENV])
		env = get_ee_from_name(blobmsg_get_string(tb[DU_UNINSTALL_ENV]));

	/* Priority should be given on EID if both EID and ENV name are provided */
	if (tb[DU_UNINSTALL_ENV_ID])
		env = get_ee_from_eeid(blobmsg_get_u32(tb[DU_UNINSTALL_ENV_ID]));

	if (env == NULL)
		return UBUS_STATUS_INVALID_ARGUMENT;

	snprintf(pkg.url, sizeof(pkg.url), "%s", blobmsg_get_string(tb[DU_UNINSTALL_NAME]));

	char output[MAX_LEN_1024] = {0};
	time_t start = time(NULL);
	int err = swmod_remove_package(env, &pkg, output, sizeof(output));

	char package_name[MAX_LEN_64], package_version[MAX_LEN_64];
	memset(package_name, 0, sizeof(package_name));
	memset(package_version, 0, sizeof(package_version));

	if (err == 0) {
		if (strstr(output, "Removing package") == NULL) {
			err = -1;
		} else {
			char *ptr = strstr(output, "Package Name:");
			if (ptr != NULL) {
				sscanf(ptr, "Package Name:%63s Version:%63s", package_name, package_version);
			}
		}
	}

	bool failed = (err) ? true : false;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);
	blobmsg_add_u8(&bb, "status", !failed);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	time_t finish = time(NULL);
	swmod_send_event_du_state_change(failed, start, finish, package_version, SWMOD_REMOVE, output);
	if (failed == false) {
		synchronize_du_list(env);
	}

	return 0;
}

static ExecEnv_t* get_ee_from_blob(struct blob_attr *msg)
{
	ExecEnv_t *env = NULL;
	struct blob_attr *tb[__EU_DU_LIST_MAX] = {NULL};

	if (blobmsg_parse(eu_du_list_policy, __EU_DU_LIST_MAX, tb, blob_data(msg), blob_len(msg)) == 0) {
		if (tb[EU_DU_LIST_ENV]) {
			env = get_ee_from_name(blobmsg_get_string(tb[EU_DU_LIST_ENV]));
		}

		/* Priority given to EID if both EID and ENV name are provided */
		if (tb[EU_DU_LIST_ENV_ID]) {
			env = get_ee_from_eeid(blobmsg_get_u32(tb[EU_DU_LIST_ENV_ID]));
		}
	}

	return env;
}

static void prepare_du_list_result(ExecEnv_t *env, struct blob_buf *bb)
{
	struct uci_section *ss = NULL;

	if (env == NULL)
		return;

	if (env->map_du_name[0] == '\0' || env->exists == false)
		return;

	swmod_uci_init(SWMOD_PATH);

	swmod_uci_foreach_section(env->map_du_name, "deployment", ss) {

		void *t = blobmsg_open_table(bb, "");

		blobmsg_add_string(bb, "du_name", swmod_uci_get_value_by_section(ss, "name"));
		blobmsg_add_string(bb, "ee_name", swmod_uci_get_value_by_section(ss, "environment"));
		blobmsg_add_string(bb, "eeid", swmod_uci_get_value_by_section(ss, "eeid"));
		blobmsg_add_string(bb, "uuid", swmod_uci_get_value_by_section(ss, "uuid"));
		blobmsg_add_string(bb, "duid", swmod_uci_get_value_by_section(ss, "duid"));
		blobmsg_add_string(bb, "url", swmod_uci_get_value_by_section(ss, "url"));
		blobmsg_add_string(bb, "version", swmod_uci_get_value_by_section(ss, "version"));
		blobmsg_add_string(bb, "config", swmod_uci_get_value_by_section(ss, "config"));
		blobmsg_add_string(bb, "description", swmod_uci_get_value_by_section(ss, "description"));
		blobmsg_add_string(bb, "vendor", ""); //TODO

		blobmsg_close_table(bb, t);
	}

	swmod_uci_fini(env->map_du_name);
}

static int
swmod_du_list(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_buf bb;
	void *a;
	ExecEnv_t *env = get_ee_from_blob(msg);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	a = blobmsg_open_array(&bb, "deployment_unit");

	if (env != NULL) {
		prepare_du_list_result(env, &bb);
	} else {
		for (int i = 0; i < MAX_ENV; i++) {
			prepare_du_list_result(&g_environments[i], &bb);
		}
	}

	blobmsg_close_array(&bb, a);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);
	return 0;
}

static void prepare_eu_list_result(ExecEnv_t *env, struct blob_buf *bb)
{
	void *t;

	if (env == NULL)
		return;

	if (env->exists == false)
		return;

	if (list_empty(&env->eu_list))
		return;

	EuNode *iter = NULL;

	list_for_each_entry(iter, &env->eu_list, list) {
		if (iter->eu.eu_exists == false) {
			continue;
		}

		t = blobmsg_open_table(bb, "");

		blobmsg_add_string(bb, "eu_name", iter->eu.name);
		blobmsg_add_string(bb, "command", iter->eu.command);
		blobmsg_add_string(bb, "state", iter->eu.state);
		blobmsg_add_string(bb, "config", iter->eu.config);
		blobmsg_add_string(bb, "version", iter->eu.version);
		blobmsg_add_string(bb, "description", iter->eu.description);
		blobmsg_add_string(bb, "ee_name", iter->eu.environment);
		blobmsg_add_u32(bb, "eeid", env->eeid);
		blobmsg_add_u32(bb, "euid", iter->eu.euid);
		blobmsg_add_u32(bb, "disk_space", iter->eu.disk_space);
		blobmsg_add_u32(bb, "memory_space", iter->eu.memory_space);
		blobmsg_add_string(bb, "vendor", iter->eu.vendor); //TODO

		blobmsg_close_table(bb, t);
	}

	swmod_delete_eu_list(&env->eu_list);
}

static int
swmod_eu_list(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_buf bb;
	void *a;

	ExecEnv_t *env = get_ee_from_blob(msg);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	a = blobmsg_open_array(&bb, "execution_unit");

	if (env != NULL) {
		populate_execution_unit(env);
		prepare_eu_list_result(env, &bb);
	} else {
		for (int i = 0; i < MAX_ENV; i++) {
			populate_execution_unit(&g_environments[i]);
			prepare_eu_list_result(&g_environments[i], &bb);
		}
	}

	blobmsg_close_array(&bb, a);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);
	return 0;
}

static int
swmod_eu_set_state(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_buf bb;
	struct blob_attr *tb[__EU_SET_STATE_SERVICE_MAX] = {NULL};
	char name[MAX_LEN_32] = {0};
	bool state = false;
	ExecEnv_t *env = NULL;

	if (blobmsg_parse(eu_set_state_policy, __EU_SET_STATE_SERVICE_MAX, tb, blob_data(msg), blob_len(msg)) != 0) {
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	if (!tb[EU_SET_STATE_SERVICE_NAME] || !tb[EU_SET_STATE_SERVICE_STATUS]) {
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	if (tb[EU_SET_STATE_SERVICE_ENV]) {
		env = get_ee_from_name(blobmsg_get_string(tb[EU_SET_STATE_SERVICE_ENV]));
	}

	/* Priority should be given on EID if both EID and ENV name are provided */
	if (tb[EU_SET_STATE_SERVICE_ENV_ID]) {
		env = get_ee_from_eeid(blobmsg_get_u32(tb[EU_SET_STATE_SERVICE_ENV_ID]));
	}

	if (env == NULL)
		return UBUS_STATUS_INVALID_ARGUMENT;

	snprintf(name, sizeof(name), "%s", blobmsg_get_string(tb[EU_SET_STATE_SERVICE_NAME]));
	state = blobmsg_get_bool(tb[EU_SET_STATE_SERVICE_STATUS]);

	int err = swmod_set_service_state(env, name, state);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_u8(&bb, "status", (err >= 0) ? true : false);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);
	return 0;
}

static int
swmod_ee_set_state(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_buf bb;
	struct blob_attr *tb[__EE_SET_STATE_MAX] = {NULL};
	ExecEnv_t *env = NULL;

	if (blobmsg_parse(ee_set_state_policy, __EE_SET_STATE_MAX, tb, blob_data(msg), blob_len(msg)) != 0) {
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	if (!tb[EE_SET_STATE_STATUS]) {
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	if (tb[EE_SET_STATE_ENV]) {
		env = get_ee_from_name(blobmsg_get_string(tb[EE_SET_STATE_ENV]));
	}

	/* Priority should be given on EID if both EID and ENV name are provided */
	if (tb[EE_SET_STATE_ENV_ID]) {
		env = get_ee_from_eeid(blobmsg_get_u32(tb[EE_SET_STATE_ENV_ID]));
	}

	if (env == NULL) {
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	char state[10] = {0};
	snprintf(state, sizeof(state), "%s", blobmsg_get_string(tb[EE_SET_STATE_STATUS]));

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	swmod_change_ee_state(env, state, &bb);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);
	return 0;
}

static const struct ubus_method swmod_object_methods[] = {
	UBUS_METHOD_NOARG("ee_list", swmod_environment),
	UBUS_METHOD("du_list", swmod_du_list, eu_du_list_policy),
	UBUS_METHOD("eu_list", swmod_eu_list, eu_du_list_policy),
	UBUS_METHOD("du_install", swmod_du_install, du_install_policy),
	UBUS_METHOD("du_update", swmod_du_update, du_update_policy),
	UBUS_METHOD("du_uninstall", swmod_du_uninstall, du_uninstall_policy),
	UBUS_METHOD("eu_set_state", swmod_eu_set_state, eu_set_state_policy),
	UBUS_METHOD("ee_set_state", swmod_ee_set_state, ee_set_state_policy),
};

static struct ubus_object_type swmod_object_type = UBUS_OBJECT_TYPE("swmodules", swmod_object_methods);

static struct ubus_object swmod_object = {
	.name = "swmodules",
	.type = &swmod_object_type,
	.methods = swmod_object_methods,
	.n_methods = ARRAY_SIZE(swmod_object_methods),
};

static int swmod_pre_init(void)
{
	char autofile[60] = {0};

	// Create lxc-auto uci file for autoboot management
	snprintf(autofile, sizeof(autofile), "%s%s", STD_UCI_PATH, LXC_AUTOBOOT_FILE);
	if (!file_exists(autofile))
		create_file(autofile);

	/* create the mount point src dir at host machine */
	create_dir(DW_DIR_PATH);

	// Populate ExecEnv details
	populate_environments();

	// Create and synchronize map_du list
	return synchronize_deployment_units_with_map_du_file();
}

static int swmod_init(struct ubus_context *ctx)
{
	int ret;

	ret = ubus_add_object(ctx, &swmod_object);
	if (ret)
		PRINT_ERR("Failed to publish '%s' object : %s", swmod_object.name, ubus_strerror(ret));

	return ret;
}

static void usage(char *prog)
{
	fprintf(stderr, "Usage: %s [options]\n", prog);
	fprintf(stderr, "\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "    -s <socket path>   ubus socket\n");
	fprintf(stderr, "    -l <log_level>     Configures log level\n");
	fprintf(stderr, "\n");
}

int main(int argc, char **argv)
{
	const char *ubus_socket = NULL;
	int ret, ch;

	while ((ch = getopt(argc, argv, "hs:l:")) != -1) {
		switch (ch) {
		case 's':
			ubus_socket = optarg;
			break;
		case 'l':
			configure_debug_level(strtol(optarg, NULL, 10));
			break;
		case 'h':
			usage(argv[0]);
			exit(0);
		default:
			break;
		}
	}

	openlog("swmodd", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	uloop_init();

	ubus_ctx = ubus_connect(ubus_socket);
	if (!ubus_ctx) {
		fprintf(stderr, "Failed to connect to ubus\n");
		return -1;
	}

	swmod_pre_init();

	ubus_add_uloop(ubus_ctx);

	ret = swmod_init(ubus_ctx);
	if (ret)
		goto out;

	uloop_run();
out:
	ubus_free(ubus_ctx);
	uloop_done();
	closelog();

	return 0;
}
