/*
 * swmod_api.c: Abstraction API to interact with host/lxc
 *
 * Copyright (C) 2021 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

/*
* Description: This file provide abstraction api to get the information from
* from host or lxc sub-system
*/

#include "opkg_utils.h"
#include "swmod_opkg.h"
#include "swmod_host.h"
#include "tools.h"  // run_cmd

#ifdef SWMOD_LXC
#include "swmod_lxc.h"
#endif

int get_pkg_list(ExecEnv_t *ee, struct list_head *pkg_list)
{
	if (ee->exists == false) {
		// delete map_du_file
		return 0;
	}

	switch (ee->ee_type) {
		case EE_TYPE_HOST:
		{
			char buffer[MAX_BUFFER_LEN] = {0};
			if (opkg_list_pkgs(buffer, sizeof(buffer)) == 0) {
				create_buffer_list(pkg_list, buffer, sizeof(buffer));
			}
		}
		break;
		case EE_TYPE_LXC:
		{
#ifdef SWMOD_LXC
			char buffer[MAX_BUFFER_LEN] = {0};
			lxc_run_cmd(ee, "opkg list", buffer, sizeof(buffer));
			create_buffer_list(pkg_list, buffer, sizeof(buffer));
#endif
		}
		break;
		default:
		{
			PRINT_ERR("Invalid ee type %s", ee->ee_type);
		}
	}
	return 0;
}

int swmod_run_cmd(ExecEnv_t *ee, char *cmd, char *output, int out_len)
{
	/* Get Config from package_name.list */
	int ret = 0;

	if (cmd == NULL || ee->exists == false) {
		PRINT_ERR("Either cmd is null or ee is disabled");
		return -1;
	}

	switch(ee->ee_type) {
		case EE_TYPE_HOST:
		{
			ret = run_cmd(cmd, output, out_len);
		}
		break;
		case EE_TYPE_LXC:
		{
#ifdef SWMOD_LXC
			ret = lxc_run_cmd(ee, cmd, output, out_len);
#endif
		}
		break;
		default:
		{
			PRINT_ERR("Invalid ee type %s", ee->ee_type);
		}
	}

	return ret;
}

int synchronize_du_list(ExecEnv_t *ee)
{
	LIST_HEAD(pkg_list);

	if (ee == NULL)
		return 0;

	if (ee->exists == false)
		return 0;

	PRINT_DEBUG("Populating DU list for %s", ee->name);
	get_pkg_list(ee, &pkg_list);
	//dump_buffer_list(&pkg_list);
	populate_opkg_deployment_unit(ee, &pkg_list);
	delete_buffer_list(&pkg_list);

	return 0;
}

int synchronize_deployment_units_with_map_du_file(void)
{
	int i = 0;

	for(i = 0;  i < MAX_ENV; i++) {
		ExecEnv_t *ee = &g_environments[i];

		if (ee->exists == false)
			continue;

		synchronize_du_list(ee);
	}

	return 0;
}

int swmod_install_package(ExecEnv_t *env, PkgInfo *pkg, char *output, int outlen)
{
	int ret = -1;

	if (output == NULL || env == NULL) {
		PRINT_ERR("env or output buffer is NULL");
		return ret;
	}

	if (pkg == NULL) {
		snprintf(output, outlen, "Internal Error");
		return ret;
	}

	if (env->exists == false) {
		snprintf(output, outlen, "ExecEnv not exists");
		return ret;
	}

	PRINT_INFO("Installing [%s] in[%s]", pkg->url, env->name);

	switch (env->ee_type) {
	case EE_TYPE_HOST:
		ret = swmod_perform_installation(pkg->url, pkg->uname, pkg->psw, output, outlen);
		break;
	case EE_TYPE_LXC:
#ifdef SWMOD_LXC
		ret = swmod_lxc_install_update_remove_package(pkg, output, outlen, env->name, SWMOD_INSTALL);
		if (strstr(output, "Package download failed") != NULL) {
			/* Download failed inside container, need to download in host */
			PRINT_WARNING("Download failed inside container, fallback to host download");
			if (0 != download_remote_package(pkg->url, pkg->uname, pkg->psw, DW_PKG_MOUNT_PATH)) {
				snprintf(output, outlen, "Package download failed from %s", pkg->url);
				return -1;
			}

			snprintf(pkg->url, sizeof(pkg->url), "%s", DW_PKG_MOUNT_PATH);
			ret = swmod_lxc_install_update_remove_package(pkg, output, outlen, env->name, SWMOD_INSTALL);

			/* delete downloaded package present in the mounted dir */
			if (file_exists(DW_PKG_MOUNT_PATH)) {
				remove(DW_PKG_MOUNT_PATH);
			}
		}
#else
		ret = -1;
#endif
		break;
	default:
		ret = -1;
	}

	return ret;
}

int swmod_update_package(ExecEnv_t *env, PkgInfo *pkg, char *output, int outlen)
{
	int ret = -1;

	if (env == NULL || output == NULL) {
		PRINT_ERR("No environment or output buffer");
		return ret;
	}

	if (pkg == NULL) {
		snprintf(output, outlen, "Internal Error");
		return ret;
	}

	if (env->exists == false) {
		snprintf(output, outlen, "ExecEnv not exists");
		return ret;
	}

	PRINT_INFO("update [%s] in[%s]", pkg->url, env->name);
	switch (env->ee_type) {
	case EE_TYPE_HOST:
		ret = swmod_perform_upgrade(pkg->url, pkg->uname, pkg->psw, output, outlen);
		break;
	case EE_TYPE_LXC:
#ifdef SWMOD_LXC
		ret = swmod_lxc_install_update_remove_package(pkg, output, outlen, env->name, SWMOD_UPDATE);
		if (strstr(output, "Package download failed") != NULL) {
			/* Download failed inside container, need to download in host */
			PRINT_WARNING("Download failed inside container, fallback to host download");
			if (0 != download_remote_package(pkg->url, pkg->uname, pkg->psw, DW_PKG_MOUNT_PATH)) {
				snprintf(output, outlen, "Package download failed from %s", pkg->url);
				return -1;
			}

			snprintf(pkg->url, sizeof(pkg->url), "%s", DW_PKG_MOUNT_PATH);
			ret = swmod_lxc_install_update_remove_package(pkg, output, outlen, env->name, SWMOD_UPDATE);

			/* delete downloaded package from mounted dir */
			if (file_exists(DW_PKG_MOUNT_PATH)) {
				remove(DW_PKG_MOUNT_PATH);
			}
		}
#else
		ret = -1;
#endif
		break;
	default:
		ret = -1;
	}

	return ret;
}

int swmod_remove_package(ExecEnv_t *env, const PkgInfo *pkg, char *output, int outlen)
{
	int ret = -1;

	if (env == NULL || output == NULL) {
		PRINT_ERR("No environment or output buffer");
		return ret;
	}

	if (pkg == NULL) {
		snprintf(output, outlen, "Internal Error");
		return ret;
	}

	if (env->exists == false) {
		snprintf(output, outlen, "ExecEnv not exists");
		return ret;
	}

	PRINT_INFO("remove [%s] from[%s]", pkg->url, env->name);
	switch (env->ee_type) {
	case EE_TYPE_HOST:
		ret = swmod_perform_uninstall(pkg->url, output, outlen);
		break;
	case EE_TYPE_LXC:
#ifdef SWMOD_LXC
		ret = swmod_lxc_install_update_remove_package(pkg, output, outlen, env->name, SWMOD_REMOVE);
#else
		ret = -1;
#endif
		break;
	default:
		ret = -1;
	}

	return ret;
}

void swmod_change_ee_state(ExecEnv_t *env, char *state, struct blob_buf *bb)
{
	if (env == NULL || state == NULL || bb == NULL) {
		blobmsg_add_u8(bb, "status", false);
		blobmsg_add_string(bb, "reason", "Insufficient information");
		return;
	}

	if (env->exists == false) {
		blobmsg_add_u8(bb, "status", false);
		blobmsg_add_string(bb, "reason", "ExecEnv not exists");
		return;
	}

	switch (env->ee_type) {
	case EE_TYPE_HOST:
		blobmsg_add_u8(bb, "status", false);
		blobmsg_add_string(bb, "reason", "request can not be performed on host system");
		break;
	case EE_TYPE_LXC:
#ifdef SWMOD_LXC
		PRINT_INFO("changing state of [%s] to [%s]", env->name, state);
		int err = swmod_set_ee_state(env, state);
		blobmsg_add_u8(bb, "status", (err == 0) ? true : false);
		blobmsg_add_string(bb, "reason", ee_state_failure_reason(err));
#else
		blobmsg_add_u8(bb, "status", false);
		blobmsg_add_string(bb, "reason", "LXC service not present");
#endif
		break;
	default:
		break;
	}

	return;
}

void populate_execution_unit(ExecEnv_t *env)
{
	if (env == NULL)
		return;

	if (env->exists == false)
		return;

	INIT_LIST_HEAD(&env->eu_list);
	switch (env->ee_type) {
	case EE_TYPE_HOST:
		/* Host system */
		get_ubus_output("service", "list", service_get_cb, env);
		break;
	case EE_TYPE_LXC:
		/* Linux containers */
#ifdef SWMOD_LXC
		populate_lxc_eu(env);
#endif
		break;
	default:
		break;
	}
}

void swmod_get_pid_detail(ExecEnv_t *env, int pid, char *cmdline, int cmd_len, int *vsize)
{
	if (env == NULL || cmdline == NULL || vsize == NULL)
		return;

	switch (env->ee_type) {
	case EE_TYPE_LXC:
#ifdef SWMOD_LXC
		get_pid_details_lxc(env, pid, cmdline, cmd_len, vsize);
#endif
		break;
	case EE_TYPE_HOST:
		get_pid_details(pid, cmdline, cmd_len, vsize);
		break;
	default:
		break;
	}
}

int swmod_set_service_state(ExecEnv_t *env, char *name, bool state)
{
	if (env == NULL) {
		PRINT_ERR("No environment info received");
		return -1;
	}

	if (env->exists == false) {
		PRINT_ERR("No environment exists");
		return -1;
	}

	if (name == NULL) {
		PRINT_ERR("No eu name found");
		return -1;
	}

	int err = -1;
	switch (env->ee_type) {
	case EE_TYPE_HOST:
		err = swmod_set_host_service_state(name, state);
		break;
	case EE_TYPE_LXC:
#ifdef SWMOD_LXC
		err = swmod_set_lxc_service_state(env, name, state);
#endif
		break;
	default:
		break;
	}

	return err;
}
