/*
 * swmod_lxc.c: SWMOD deamon
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#include <stdio.h>
#include <unistd.h>
#include <sys/sysinfo.h>
#include <sys/statvfs.h>
#include <sys/utsname.h>
#include <errno.h>
#include <fcntl.h>

#include "tools.h"
#include "swmod_uci.h"
#include "swmod_opkg.h"
#include "swmod.h"
#include "swmod_lxc.h"
#include "swmod_api.h"

#define LXC_CONFIG_PATH "/etc/lxc/lxc.conf"
#define LXC_PATH "lxc.lxcpath"

static char g_lxc_buff[MAX_BUFFER_LEN];

enum {
	LXC_START,
	LXC_STOP,
	LXC_PAUSE,
	LXC_RESUME,
	__LXC_STATE_MAX
};

enum {
	LXC_SUCCESS,
	LXC_FAILURE,
	LXC_NOT_SUPPORTED,
	LXC_ID_NOT_FOUND,
	LXC_ALREADY_RUNNING,
	LXC_NOT_RUNNING,
	INVALID_REQUEST,
	__LXC_ERROR_MAX
};

static int get_requested_lxc_state(char *state)
{
	if (strcmp(state, "start") == 0) {
		return LXC_START;
	} else if (strcmp(state, "stop") == 0) {
		return LXC_STOP;
	} else if (strcmp(state, "pause") == 0) {
		return LXC_PAUSE;
	} else if (strcmp(state, "resume") == 0) {
		return LXC_RESUME;
	} else {
		return __LXC_STATE_MAX;
	}
}

char ee_set_state_error_msg[__LXC_ERROR_MAX][MAX_LEN_64] = {
	"",
	"internal failure",
	"LXC not supported",
	"container not found",
	"ExecEnv already running",
	"ExecEnv not running",
	"invalid request"
};

typedef struct lxc_attach_args {
	const char *lxcpath;
	const char *value;
	int action;
	int pid;
	int fd;
	const char *user;
	const char *pwd;
} lxc_attach_args;

struct service_state {
	char *name;
	bool state;
};

const char *ee_state_failure_reason(int err)
{
	if (err >= __LXC_ERROR_MAX || err < 0)
		return "unknown";

	return ee_set_state_error_msg[err];
}

const char *get_lxc_path_from_config(void)
{
	return lxc_get_global_config_item(LXC_PATH);
}

bool lxc_is_supported(const char **lxcpath)
{
	if (!file_exists(LXC_CONFIG_PATH))
		return false;

	*lxcpath = get_lxc_path_from_config();

	if (!(*lxcpath) || !file_exists(*lxcpath))
		return false;

	return true;
}

static int lxc_attach_func(struct lxc_container *ct, lxc_attach_exec_t exec_function,
			   lxc_attach_args *command, bool get_output)
{
	int ret;
	pid_t pid;
	int pipefd[2];
	lxc_attach_options_t attach_options = LXC_ATTACH_OPTIONS_DEFAULT;
	lxc_attach_args lxc_input;

	memset(&lxc_input, 0, sizeof(lxc_attach_args));

	g_lxc_buff[0] = '\0';
	ret = pipe2(pipefd, O_NONBLOCK);
	if (ret < 0) {
		PRINT_ERR("Failed to create pipe");
		return -1;
	}

	if (command) {
		lxc_input.lxcpath = command->lxcpath;
		lxc_input.value = command->value;
		lxc_input.action = command->action;
		lxc_input.pid = command->pid;
		lxc_input.user = command->user;
		lxc_input.pwd = command->pwd;
	}
	lxc_input.fd = pipefd[1];

	const char *state = ct->state(ct);

	if (strcmp(state, "RUNNING") != 0) {
		PRINT_ERR("Container might be frozen");
		goto err;
	}

	ret = ct->attach(ct, exec_function, &lxc_input, &attach_options, &pid);
	if (ret < 0)
		goto err;

	wait_for_pid(pid);

	if (get_output == true) {
		ret = read(pipefd[0], &g_lxc_buff, sizeof(g_lxc_buff)-1);
		if (ret != -1) { // error on read
			g_lxc_buff[ret] = '\0';
		}
		PRINT_DEBUG("== Attach Result buff [%s] ##", g_lxc_buff);
	}

err:
	close(pipefd[0]);
	close(pipefd[1]);
	return ret;
}

static void dump_output_fd(int fd, const char *str)
{
	if (str == NULL)
		return;

	write(fd, str, strlen(str));
	close(fd);
}

/************************** Execution Environments **************************/
static int lxc_attach_run_env_func(void *args)
{
	char buff[MAX_LEN_1024];
	struct utsname utsname;
	char vendor[MAX_LEN_128] = {0}, version[MAX_LEN_16] = {0};
	lxc_attach_args *data = (lxc_attach_args *) args;

	if (uname(&utsname) >= 0) {
		swmod_strncpy(vendor, utsname.nodename, MAX_LEN_128);
		swmod_strncpy(version, utsname.release, MAX_LEN_16);
	}

	struct sysinfo sinfo;
	unsigned long alloc_mem = 0, avail_mem = 0;

	if (sysinfo(&sinfo) == 0) {
		alloc_mem = (sinfo.totalram / 1024);
		avail_mem = (sinfo.freeram / 1024);
	}

	struct statvfs dinfo;
	unsigned long allocated_disk_space = 0;
	unsigned long available_disk_space = 0;

	if (statvfs("/", &dinfo) == 0) {
		allocated_disk_space = (dinfo.f_bsize * dinfo.f_blocks) / 1024;
		available_disk_space = (dinfo.f_bsize * dinfo.f_bfree) / 1024;
	}

	/* lxc_attach_result buffer format */
	/* type=<ENV_TYPE> vendor=<ENV_VENDOR> version=<ENV_VERSION>
	 *  alloc_mem=<ENV_ALLOCATED_MEMORY> avail_mem=<ENV_AVAILABLE_MEMORY> */
	snprintf(buff, sizeof(buff), "vendor=%s version=%s alloc_mem=%lu avail_mem=%lu alloc_disk=%lu avail_disk=%lu",
			vendor, version, alloc_mem, avail_mem, allocated_disk_space, available_disk_space);

	dump_output_fd(data->fd, buff);

	return 0;
}

unsigned long long lxc_get_cgroup_item_ull(struct lxc_container *ct, const char *config)
{
	unsigned long long value = 0;
	char buffer[MAX_LEN_32] = {0};

	int res = ct->get_cgroup_item(ct, config, buffer, sizeof(buffer));
	if (res >= 0) {
		sscanf(buffer, "%llu", &value);
	}
	return value;
}

void populate_lxc_environment(void)
{
	const char *lxcpath = NULL;

	if (!lxc_is_supported(&lxcpath))
		return;

	struct lxc_container **clist = NULL;
	int i, lxc_nbr;

	lxc_nbr = list_all_containers(lxcpath, NULL, &clist);
	if (clist == NULL || lxc_nbr == 0) {
		PRINT_INFO("Empty container list, return");
		return;
	}

	for (i = 0; i < lxc_nbr && i < (MAX_ENV - LXC_START_INDEX); i++) {
		struct lxc_container *ct = clist[i];
		bool ct_running = ct->is_running(ct);
		char buffer[MAX_LEN_1024];
		ExecEnv_t *ee;

		ee = &g_environments[LXC_START_INDEX + i];
		ee->exists = true;
		ee->ee_type = EE_TYPE_LXC;
		ee->eeid = LXC_START_INDEX + i + 1;
		ee->parent_eeid = LXC_START_INDEX;

		swmod_strncpy(ee->type, "Linux Container", MAX_LEN_64);
		swmod_strncpy(ee->name, ct->name, MAX_LEN_32);
		snprintf(ee->map_du_name, MAX_LEN_64, "%s_%s", SWMOD_MAP_DU, ct->name);
		ee->autoboot = get_autoboot_from_config_file(ct->name);
		swmod_strncpy(ee->status, ct_running ? "Up" : "Disabled", MAX_LEN_32);

		if (ct_running && ((strcmp(ct->state(ct), "FROZEN") == 0) ||
			strcmp(ct->state(ct), "FREEZING") == 0)) {
				ee->pause = 1;
		}

		if (!ct_running || ee->pause) {
			PRINT_INFO("lxc container not running or frozen");
			lxc_container_put(ct);
			continue;
		}

		snprintf(buffer, MAX_LEN_1024, "%s%s", SWMOD_PATH, ee->map_du_name);
		if (file_exists(buffer) == false) {
			create_file(buffer);
			synchronize_du_list(ee);
		}

		/* get memory space limit in cgroup */
		unsigned long long val = 0;
		val = lxc_get_cgroup_item_ull(ct, "memory.limit_in_bytes");
		ee->allocated_memory = val / 1024;

		/* get free space from lxc info */
		val = lxc_get_cgroup_item_ull(ct, "memory.usage_in_bytes");
		ee->available_memory = ee->allocated_memory - val / 1024;

		int ret = lxc_attach_func(ct, lxc_attach_run_env_func, NULL, true);
		if (ret >= 0 && strlen(g_lxc_buff)) {
				sscanf(g_lxc_buff, "vendor=%127s version=%15s alloc_mem=%lu avail_mem=%lu alloc_disk=%lu avail_disk=%lu",
					ee->vendor, ee->version, &ee->allocated_memory, &ee->available_memory,
					&ee->allocated_disk_space, &ee->available_disk_space);
		}

		if (ret == -1) {
			swmod_strncpy(ee->status, "Error", MAX_LEN_32);
		}

		lxc_container_put(ct);
	}

	if (lxc_nbr > 0)
		FREE(clist);
}

void lxc_service_get_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	char *str;
	lxc_attach_args *input = (lxc_attach_args *) req->priv;

	str = blobmsg_format_json(msg, true);
	if (str != NULL) {
		write(input->fd, str, strlen(str));
		free(str);
	}
}

static int lxc_attach_pid(void *args)
{
	char buff[MAX_LEN_1024];
	char cmdline[MAX_LEN_32] = {0};
	int vsize = 0;
	lxc_attach_args *lxc_input = (lxc_attach_args *) args;

	get_pid_details(lxc_input->pid, cmdline, MAX_LEN_32, &vsize);
	/* lxc_attach_result buffer format */
	/* type=<ENV_TYPE> vendor=<ENV_VENDOR> version=<ENV_VERSION>
	 *  alloc_mem=<ENV_ALLOCATED_MEMORY> avail_mem=<ENV_AVAILABLE_MEMORY> */
	snprintf(buff, sizeof(buff), "cmdline=%s vsize=%d", cmdline, vsize);

	dump_output_fd(lxc_input->fd, buff);

	return 0;
}

void get_pid_details_lxc(ExecEnv_t *env, int pid, char *cmdline, int cmd_len, int *vsize)
{
	if (env == NULL || cmdline == NULL || cmd_len < MAX_LEN_32 || vsize == NULL)
		return;

	if (env->name[0] == '\0')
		return;

	struct lxc_container *ct = lxc_container_new(env->name, NULL);
	if (!ct) {
		PRINT_ERR("Failed in opening container: %s", env->name);
		return;
	}

	if (!ct->is_running(ct)) {
		lxc_container_put(ct);
		return;
	}

	lxc_attach_args lxc_data;

	memset(&lxc_data, 0, sizeof(lxc_attach_args));
	lxc_data.pid = pid;
	int ret = lxc_attach_func(ct, lxc_attach_pid, &lxc_data, true);
	if (ret >= 0 && strlen(g_lxc_buff)) {
		sscanf(g_lxc_buff, "cmdline=%31s vsize=%d", cmdline, vsize);
	}

	lxc_container_put(ct);
}

static int lxc_attach_eu_list(void *args)
{
	get_ubus_output("service", "list", lxc_service_get_cb, args);
	return 0;
}

void populate_lxc_eu(ExecEnv_t *env)
{
	struct blob_buf bb;

	if (env == NULL)
		return;

	if (env->name[0] == '\0' || env->exists == false)
		return;

	struct lxc_container *ct = lxc_container_new(env->name, NULL);
	if (!ct) {
		PRINT_ERR("Failed in opening container: %s", env->name);
		return;
	}

	if (!ct->is_running(ct)) {
		lxc_container_put(ct);
		return;
	}

	int ret = lxc_attach_func(ct, lxc_attach_eu_list, NULL, true);
	if (ret >= 0 && strlen(g_lxc_buff)) {
		memset(&bb, 0, sizeof(struct blob_buf));
		blob_buf_init(&bb, 0);
		blobmsg_add_json_from_string(&bb, g_lxc_buff);
		update_eu_from_blob(env, bb.head);
		blob_buf_free(&bb);
	}

	lxc_container_put(ct);
}

/************************** Deployment/Execution Unit **************************/
/************************** Install/Remove **************************/
static int lxc_attach_run_opkg_method_func(void *args)
{
	lxc_attach_args *data = (lxc_attach_args *)args;

	if (!data || *(data->value) == '\0') {
		PRINT_ERR("No data available");
		return -1;
	}

	int err = -1;

	char output[MAX_LEN_1024] = {0};
	if (data->action == SWMOD_INSTALL)
		err = swmod_perform_installation(data->value, data->user, data->pwd, output, sizeof(output));
	else if (data->action == SWMOD_UPDATE)
		err = swmod_perform_upgrade(data->value, data->user, data->pwd, output, sizeof(output));
	else if (data->action == SWMOD_REMOVE)
		err = swmod_perform_uninstall(data->value, output, sizeof(output));
	else
		err = -1;

	dump_output_fd(data->fd, output);
	return err;
}

int swmod_lxc_install_update_remove_package(const PkgInfo *pkg, char *out, int outlen, const char *ct_name, int action)
{
	int err = -1;
	struct lxc_container *ct = NULL;

	if (out == NULL) {
		PRINT_ERR("No output buffer");
		return err;
	}

	if (pkg == NULL) {
		snprintf(out, outlen, "Internal Error");
		PRINT_ERR("No package information");
		return err;
	}

	if (ct_name == NULL) {
		snprintf(out, outlen, "No ExecEnv information received");
		PRINT_ERR("No environment information given");
		return err;
	}

	if (ct_name[0] == '\0') {
		snprintf(out, outlen, "ExecEnv name unknown");
		PRINT_ERR("No environment name given");
		return err;
	}

	ct = lxc_container_new(ct_name, NULL);
	if (!ct) {
		snprintf(out, outlen, "Error in opening container: %s", ct_name);
		PRINT_ERR("Failed in opening container: %s", ct_name);
		return err;
	}

	if (!ct->is_running(ct)) {
		snprintf(out, outlen, "ExecEnv not running");
		lxc_container_put(ct);
		return err;
	}

	PRINT_DEBUG("## Container %s", ct->name);
	lxc_attach_args lxc_args = (lxc_attach_args){.value = pkg->url, .action = action, .user = pkg->uname, .pwd = pkg->psw};
	err = lxc_attach_func(ct, lxc_attach_run_opkg_method_func, &lxc_args, true);
	PRINT_DEBUG("# Reading i/u/r %d, lxc[%s], err %d", action, ct->name, err);
	if (err >= 0  && strlen(g_lxc_buff)) {
		err = 0;
		snprintf(out, outlen, "%s", g_lxc_buff);
	}

	lxc_container_put(ct);

	return err;
}

static int lxc_attach_service_state_func(void *args)
{
	struct service_state *st = (struct service_state *)args;

	return ubus_call_service_state(st->name, st->state);
}

int swmod_set_ee_state(ExecEnv_t *env, char *state)
{
	int timeout;
	int ret = LXC_SUCCESS;

	struct lxc_container *ct = NULL;
	if (env == NULL) {
		ret = INVALID_REQUEST;
		PRINT_ERR("No environment information given");
		return ret;
	}

	if (env->name[0] == '\0'|| state == NULL) {
		ret = INVALID_REQUEST;
		PRINT_ERR("No environment information or state given");
		return ret;
	}

	ct = lxc_container_new(env->name, NULL);
	if (!ct) {
		ret = LXC_ID_NOT_FOUND;
		PRINT_ERR("Failed in opening container: %s", env->name);
		return ret;
	}

	int req_state = get_requested_lxc_state(state);
	bool need_du_sync = false;

	switch (req_state) {
	case LXC_START:
		if (!get_autoboot_from_config_file(ct->name))
			set_autoboot_to_config_file(ct->name);

		if (ct->is_running(ct)) {
			need_du_sync = true;
			ret = LXC_ALREADY_RUNNING;
			break;
		}

		ct->want_close_all_fds(ct, true);
		ret = ct->start(ct, 0, NULL) ? LXC_SUCCESS: LXC_FAILURE;
		if (ret == LXC_SUCCESS) {
			need_du_sync = true;
		}
		break;
	case LXC_STOP:
		timeout = get_timeout_from_config_file(ct->name);

		if (get_autoboot_from_config_file(ct->name))
			delete_autoboot_from_config_file(ct->name);

		if (!ct->is_running(ct)) {
			ret = LXC_NOT_RUNNING;
			break;
		}

		if (strcmp(ct->state(ct), "FROZEN") == 0 || strcmp(ct->state(ct), "FREEZING") == 0) {
			/* to stop a frozen container first need to unfreeze it */
			if (!ct->unfreeze(ct)) {
				ret = LXC_FAILURE;
				break;
			}
		}

		ret = ct->shutdown(ct, timeout) ? LXC_SUCCESS : LXC_FAILURE;
		if (ret == LXC_FAILURE) {
			PRINT_DEBUG("Going to force stop [%s]", ct->name);
			ret = ct->stop(ct) ? LXC_SUCCESS : LXC_FAILURE;
		}
		break;
	case LXC_PAUSE:
		if (!ct->is_running(ct)) {
			ret = LXC_NOT_RUNNING;
			break;
		}

		ret = ct->freeze(ct) ? LXC_SUCCESS : LXC_FAILURE;
		break;
	case LXC_RESUME:
		if (!ct->is_running(ct)) {
			ret = LXC_NOT_RUNNING;
			break;
		}

		ret = ct->unfreeze(ct) ? LXC_SUCCESS : LXC_FAILURE;
		break;
	default:
		ret = INVALID_REQUEST;
	}

	if (need_du_sync == true) {
		char buffer[MAX_LEN_1024];
		snprintf(buffer, MAX_LEN_1024, "%s%s", SWMOD_PATH, env->map_du_name);
		if (file_exists(buffer) == false) {
			create_file(buffer);
		}
		synchronize_du_list(env);
	}

	lxc_container_put(ct);

	return ret;
}

int swmod_set_lxc_service_state(ExecEnv_t *env, char *name, bool state)
{
	int ret = -1;
	struct lxc_container *ct = NULL;

	if (env == NULL) {
		PRINT_ERR("No environment info");
		return ret;
	}

	if (env->name[0] == '\0') {
		PRINT_ERR("No ExecEnv name found");
		return ret;
	}

	if (name == NULL) {
		PRINT_ERR("No eu name information");
		return ret;
	}

	ct = lxc_container_new(env->name, NULL);
	if (!ct) {
		PRINT_ERR("Failed in opening container: %s", env->name);
		return ret;
	}

	if (!ct->is_running(ct)) {
		PRINT_INFO("lxc container not running");
		lxc_container_put(ct);
		return ret;
	}

	struct service_state st;
	st.name = name;
	st.state = state;
	ret = lxc_attach_func(ct, lxc_attach_service_state_func, (lxc_attach_args *)&st, false);
	lxc_container_put(ct);

	return ret;
}

static int lxc_attach_run_cmd(void *args)
{
	lxc_attach_args *lxc_input = (lxc_attach_args *) args;
	FILE *pp; // program pointer
	const char *cmd;

	cmd = lxc_input->value;
	if (cmd == NULL) { // null command to run, silently ignore
		PRINT_ERR("Empty command to run");
		return 0;
	}

	pp = popen(cmd, "r");
	if (pp != NULL) {
		char line[MAX_LEN_512] = {0};

		while(fgets(line, sizeof(line), pp) != NULL) {
			write(lxc_input->fd, line, strlen(line));
		}
		pclose(pp);
	}

	return 0;
}

int lxc_run_cmd(ExecEnv_t *ee, const char *cmd, char *output, int out_len)
{
	struct lxc_container *ct;
	lxc_attach_args lxc_data;

	if (ee->exists == false || ee->ee_type != EE_TYPE_LXC) {
		PRINT_ERR("lxc not enabled");
		return -1;
	}

	memset(&lxc_data, 0, sizeof(lxc_attach_args));
	lxc_data.value = cmd;
	ct = lxc_container_new(ee->name, NULL);
	if (!ct) {
		PRINT_ERR("failed to open container");
		return -1;
	}

	if (!ct->is_running(ct)) {
		PRINT_ERR("container not running");
		lxc_container_put(ct);
		return -1;
	}

	int ret = lxc_attach_func(ct, lxc_attach_run_cmd, &lxc_data, true);
	if (ret >= 0 && strlen(g_lxc_buff)) {
		snprintf(output, out_len, "%s", g_lxc_buff);
	}

	lxc_container_put(ct);

	return ret;
}

