/*
 * tools.c: SWMOD deamon
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#include <stdlib.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <uuid/uuid.h>
#include <stdarg.h>
#include <libubus.h>
#include <json-c/json.h>
#include <libubox/blobmsg_json.h>

#include "swmod_uci.h"
#include "swmod_opkg.h"
#include "opkg_utils.h"
#include "swmod_host.h"
#include "tools.h"
#include "swmod_api.h"

#define PROC_PATH "/proc"
#define DEFAULT_LOG_LEVEL (LOG_INFO)

static unsigned char gLogLevel = DEFAULT_LOG_LEVEL;

// Logging utilities
void configure_debug_level(unsigned char level)
{
	if (level > LOG_DEBUG) {
		fprintf(stderr, "Configured log level %d invalid, setting default level", level);
		gLogLevel = DEFAULT_LOG_LEVEL;
	} else {
		gLogLevel = level;
	}
}

void print_error(const char *format, ...)
{
	va_list arglist;

	if (gLogLevel < LOG_ERR)
		return;

	va_start(arglist, format);
	vsyslog(LOG_ERR, format, arglist);
	va_end(arglist);
}

void print_warning(const char *format, ...)
{
	va_list arglist;

	if (gLogLevel < LOG_WARNING)
		return;

	va_start(arglist, format);
	vsyslog(LOG_WARNING, format, arglist);
	va_end(arglist);
}

void print_info(const char *format, ...)
{
	va_list arglist;

	if (gLogLevel < LOG_INFO)
		return;

	va_start(arglist, format);
	vsyslog(LOG_INFO, format, arglist);
	va_end(arglist);
}

void print_debug(const char *format, ...)
{
	va_list arglist;

	if (gLogLevel < LOG_DEBUG)
		return;

	va_start(arglist, format);
	vsyslog(LOG_DEBUG, format, arglist);
	va_end(arglist);
}

bool file_exists(const char *path)
{
	struct stat statbuf;

	return stat(path, &statbuf) == 0;
}

void remove_new_line(char *buf)
{
	int len;

	if (buf == NULL)
		return;

	len = strlen(buf) - 1;
	if (buf[len] == '\n')
		buf[len] = 0;
}

void create_file(const char *path)
{
	if (!file_exists(path)) {
		/*
		 *File does not exist
		 */
		FILE *fp = fopen(path, "w"); // new empty file
		if (fp) {
			fclose(fp);
		}
		sync();
	}
}

bool dir_exist(const char *dir)
{
	struct stat sb;

	if (stat(dir, &sb) == 0 && S_ISDIR(sb.st_mode)) {
		return true;
	}

	return false;
}

bool create_dir(const char *path)
{
	if (dir_exist(path) == true)
		return true;

	/* not present so create new */
	if (0 != mkdir(path, 0700))
		return false;

	return true;
}

void swmod_strncpy(char *dst, const char *src, size_t n)
{
	strncpy(dst, src, n - 1);
	dst[n - 1] = 0;
}

char *generate_uuid(void)
{
	uuid_t binuuid;

	uuid_generate_random(binuuid);
	char *uuid = malloc(37);
	uuid_unparse(binuuid, uuid);

	return uuid;
}

char *generate_duid(bool sysnchronise, int number)
{
	const char buf[] = "0123456789abcdefghijklmnopqrstuvwxyz";
	int div = sizeof(buf) - 1;
	char euid[16] = {0};
	int i;

	srand(time(NULL));
	if (sysnchronise) {
		char euid_num[8] = {0};

		for (i = 0; i < 3; i++)
			euid[i] = buf[rand() % div];
		snprintf(euid_num, sizeof(euid_num), "%04d", number);
		strncat(euid, euid_num, 4);
	} else {
		for (i = 0; i < 7; i++)
			euid[i] = buf[rand() % div];
	}
	euid[7] = '\0';

	return strdup(euid);
}

/*
 * Description: Run a system command using popen
 *
 * input:
 * // cmd: command to run
 *
 * ouput:
 * // ouput: output of the executed command will be returned to output buffer
 * // out_len: Max length of output buffer
 *
 * returns:
 * // -1: In case of faults
 * // 0: In case of success
 */
int run_cmd(const char *cmd, char *output, size_t out_len)
{
	int ret = -1;
	FILE *pp; // program pointer

	if (cmd == NULL) // null command to run, silently ignore
		return 0;

	// init the buffer
	if (output != NULL) {
		output[0] = '\0';
	}

	pp = popen(cmd, "r");
	if (pp != NULL) {
		char line[512] = {0};

		if (output == NULL || out_len == 0) { // output not required
			return 0;
		}

		while(fgets(line, sizeof(line), pp) != NULL) {
			int line_len = strlen(line);
			if (strlen(output) + line_len < out_len) {
				strncat(output, line, line_len);
			}
		}
		pclose(pp);
		ret = 0;
	}

	return ret;
}

void delete_autoboot_from_config_file(const char *env)
{
	if (0 != swmod_uci_init(STD_UCI_PATH))
		return;

	struct uci_section *s = NULL, *stmp = NULL;
	swmod_uci_foreach_section_safe(LXC_AUTOBOOT_FILE, "container", stmp, s) {
		const char *name = swmod_uci_get_value_by_section(s, "name");

		/* Check name exists, if yes => remove the section */
		if (strcmp(name, env) == 0) {
			swmod_uci_delete_by_section(s, NULL, NULL);
			break;
		}
	}

	swmod_uci_fini(LXC_AUTOBOOT_FILE);
}

void set_autoboot_to_config_file(const char *env)
{
	if (0 != swmod_uci_init(STD_UCI_PATH))
		return;

	struct uci_section *new_s = swmod_uci_add_section(LXC_AUTOBOOT_FILE, "container");

	swmod_uci_set_value_by_section(new_s, "name", env);
	swmod_uci_set_value_by_section(new_s, "timeout", "300");

	swmod_uci_fini(LXC_AUTOBOOT_FILE);
}

bool get_autoboot_from_config_file(const char *env)
{
	bool found = false;

	if (swmod_uci_init(STD_UCI_PATH) != 0)
		return found;

	struct uci_section *s = NULL;
	swmod_uci_foreach_section(LXC_AUTOBOOT_FILE, "container", s) {
		const char *name = swmod_uci_get_value_by_section(s, "name");
		if (strcmp(name, env) == 0) {
			found = true;
			break;
		}
	}

	swmod_uci_fini(LXC_AUTOBOOT_FILE);
	return found;
}

bool present_in_buffer_list(struct list_head *plist, char *entry)
{
	struct buffer_list *pos;

	if (entry == NULL)
		return false;

	list_for_each_entry(pos, plist, list) {
		if (strstr(pos->line, entry)) {
			return true;
		}
	}

	return false;
}

void buffer_add_line(struct list_head *head, char *entry)
{
	struct buffer_list *node = NULL;

	if (entry == NULL)
		return;

	node = (struct buffer_list *) malloc(sizeof(*node));
	if (!node) {
		PRINT_ERR("Out of memory!");
		return;
	}

	swmod_strncpy(node->line, entry, MAX_LEN_128);

	INIT_LIST_HEAD(&node->list);
	list_add_tail(&node->list, head);
}

void create_buffer_list(struct list_head *opkg_list, char *buffer, int buff_len)
{
	char *token = NULL;

	token = strtok(buffer, "\n");
	while (token != NULL) {
		buffer_add_line(opkg_list, token);
		token = strtok(NULL, "\n");
	}
}

void delete_buffer_list(struct list_head *opkg_list)
{
	struct buffer_list *iter, *node;

	list_for_each_entry_safe (iter, node, opkg_list, list) {
		list_del(&iter->list);
		free(iter);
	}
}

int get_pid_details(int pid, char *cmdline, int cmd_len, int *vsize)
{
	char fcmd[512];
	char fstat[512];
	unsigned int bsize = 0;
	int spid, ppid;
	FILE *fp;

	if (cmdline == NULL || vsize == NULL || cmd_len < MAX_LEN_32)
		return 0;

	snprintf(fcmd, sizeof(fcmd), "%s/%d/cmdline", PROC_PATH, pid);

	*vsize = 0;
	strncpy(cmdline, "", cmd_len);

	fp = fopen(fcmd, "r");
	if (fp == NULL)
		return 0;

	if (fgets(cmdline , 31 , fp) == NULL) {
		fclose(fp);
		return 0;
	}
	fclose(fp);

	snprintf(fstat, sizeof(fstat), "%s/%d/stat", PROC_PATH, pid);
	if (file_exists(fstat) == false)
		return 0;

	fp = fopen(fstat, "r");
	if (fp == NULL)
		return 0;

	int ret = fscanf(fp, "%d %*s %*c %d %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %u",
			&spid,
			&ppid,
			&bsize);

	if (ret != 3) {
		fclose(fp);
		return 0;
	}

	fclose(fp);
	*vsize = bsize / 1024;

	return 0;
}

int wait_for_pid(pid_t pid)
{
	int status, ret;

again:
	ret = waitpid(pid, &status, 0);
	if (ret == -1) {
		if (errno == EINTR)
			goto again;

		return -1;
	}

	if (ret != pid)
		goto again;

	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
		return -1;

	return 0;
}

void update_eu_from_blob(ExecEnv_t *env, struct blob_attr *msg)
{
	struct blob_attr *service, *instances, *inst;
	size_t rem, rem2, rem3;
	enum {
		P_RUN,
		P_PID,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "running", BLOBMSG_TYPE_BOOL },
		{ "pid", BLOBMSG_TYPE_INT32 },
	};

	if (env == NULL) {
		PRINT_ERR("Environment info not provided");
		return;
	}

	if (env->exists == false) {
		PRINT_ERR("Environment not exists");
		return;
	}

	blobmsg_for_each_attr(service, msg, rem) {
		if (is_compulsory_service(blobmsg_name(service)) == true) {
			continue;
		}

		blobmsg_for_each_attr(instances, service, rem2) {
			ExecUnit new;
			memset(&new, 0, sizeof(ExecUnit));
			new.eu_exists = true;
			swmod_strncpy(new.name, blobmsg_name(service), sizeof(new.name));
			swmod_strncpy(new.environment, env->name, sizeof(new.environment));

			blobmsg_for_each_attr(inst, instances, rem3) {
				struct blob_attr *tb[__P_MAX] = {NULL};

				if (blobmsg_parse(p, __P_MAX, tb, blobmsg_data(inst), blobmsg_len(inst)) != 0)
					continue;

				if (tb[P_PID]) {
					new.euid = blobmsg_get_u32(tb[P_PID]);
					swmod_get_pid_detail(env, new.euid, new.command, sizeof(new.command), &new.memory_space);
				} else {
					new.euid = -1;
				}

				if (tb[P_RUN] && blobmsg_get_bool(tb[P_RUN])) {
					swmod_strncpy(new.state, "Active", sizeof(new.state));
				} else {
					swmod_strncpy(new.state, "Idle", sizeof(new.state));
				}

				swmod_strncpy(new.vendor, "", sizeof(new.vendor));  // TODO
				break;
			}
			swmod_add_eu_in_list(&env->eu_list, &new);
		}
	}
	/*
	swmod_strncpy(exec_units[0].version[nbr], version, 32);
	exec_units[0].disk_space[nbr] = disk_space;
	swmod_strncpy(exec_units[0].description[nbr], spch+14, 1024);
	swmod_strncpy(exec_units[0].config[nbr], spch+12, 32);
	*/
}

void get_ubus_output(char *object, char *method, ubus_data_handler_t callback, void *data)
{
	/* This function can be called inside LXC, so need a new ubux ctx */
	uint32_t id;
	int fault = 0;
	struct ubus_context *ctx = ubus_connect(NULL);
	struct blob_buf bb;

	fault = ubus_lookup_id(ctx, object, &id);
	if (fault) {
		PRINT_ERR("ubus service not found %d", fault);
		ubus_free(ctx);
		return;
	}

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	// Invoke Ubus to get data from uspd
	fault = ubus_invoke(ctx, id, method, bb.head, callback, data, UBUS_TIMEOUT);

	blob_buf_free(&bb);
	ubus_free(ctx);
}

int ubus_call_service_state(char *service, bool state)
{
	uint32_t id;
	int rc = -1;

	/* This function can be called inside LXC, so need a new ubux ctx */
	struct ubus_context *ctx = ubus_connect(NULL);
	if (ctx == NULL) {
		return rc;
	}

	struct blob_buf b;
	memset(&b, 0, sizeof(struct blob_buf));

	blob_buf_init(&b, 0);
	blobmsg_add_u8(&b, "spawn", state);
	blobmsg_add_string(&b, "name", service);

	if (!ubus_lookup_id(ctx, "service", &id))
		rc = ubus_invoke(ctx, id, "state", b.head, NULL, NULL, 5);

	blob_buf_free(&b);
	ubus_free(ctx);

	return rc;
}

int get_timeout_from_config_file(const char *env)
{
	long timeout = 300; /* default timeout */

	if (swmod_uci_init(STD_UCI_PATH) != 0) {
		return timeout;
	}

	struct uci_section *s = NULL;
	swmod_uci_foreach_section(LXC_AUTOBOOT_FILE, "container", s) {
		const char *name = swmod_uci_get_value_by_section(s, "name");
		if (strcmp(name, env) == 0) {
			const char *time = swmod_uci_get_value_by_section(s, "timeout");
			if (time != NULL && time[0] != '\0') {
				timeout = strtol(time, NULL, 10);
			}
			break;
		}
	}

	swmod_uci_fini(LXC_AUTOBOOT_FILE);
	return timeout;
}

void swmod_add_eu_in_list(struct list_head *eu_list, ExecUnit *eu)
{
	if (eu == NULL) {
		return;
	}

	EuNode *node = (EuNode *)malloc(sizeof(EuNode));
	if (node == NULL) {
		PRINT_ERR("Out of memory");
		return;
	}

	memcpy(&node->eu, eu, sizeof(ExecUnit));
	INIT_LIST_HEAD(&node->list);
	list_add_tail(&node->list, eu_list);
}

void swmod_delete_eu_list(struct list_head *eu_list)
{
	EuNode *iter = NULL, *node;

	list_for_each_entry_safe (iter, node, eu_list, list) {
		list_del(&iter->list);
		free(iter);
	}
}
